from __future__ import annotations

import inspect
import keyword
import os
import re
from collections import defaultdict, deque, OrderedDict as ColOrderedDict, namedtuple
from enum import Enum
from sqlalchemy import Enum as SqlEnum, ARRAY
from typing import (
    Type,
    Any,
    List,
    Tuple,
    Dict,
    Optional,
    DefaultDict,
    Set,
    FrozenSet,
    Deque,
    OrderedDict,
    Union,
)

import inflect
from sqlalchemy.sql.type_api import TypeEngine

DEFAULT_INDENT_CHAR = os.getenv("DEFAULT_IDENT_CHAR", " ")
DEFAULT_INDENT_MULTIPLIER = int(os.getenv("DEFAULT_IDENT_MULTIPLIER", 4))

DEFAULT_INDENTATION = DEFAULT_INDENT_CHAR * DEFAULT_INDENT_MULTIPLIER


class Empty:
    pass


empty = Empty()


class InflectEngine:
    def __init__(self):
        self.engine = inflect.engine()

    def to_singular(self, word: str) -> str:
        singular = self.engine.singular_noun(word)
        return singular if singular else word

    def to_plural(self, word: str) -> str:
        # HACK! plural_noun doesn't work properly if word is already plural
        plural = self.engine.plural_noun(self.to_singular(word))
        return plural if plural else word


inflect_engine = InflectEngine()


class StringReprWrapper:
    __slots__ = ("wrapped",)

    def __init__(self, wrapped: str):
        self.wrapped = wrapped

    def __repr__(self):
        return self.wrapped

    def __eq__(self, other):
        return self.wrapped == other.wrapped


class ImportParts:
    __slots__ = ("module", "main_class", "inner")

    def __init__(self, obj: Any):
        module, maybe_class = self.get_module_and_class(obj)
        main_class = ""
        inner = ""
        if isinstance(obj, str):
            module = obj

        elif module in ("builtins", "__main__"):
            module = ""
            main_class = maybe_class.__name__

        elif maybe_class is not empty:
            main_class, inner = self.parse_qualified_name(maybe_class)

        self.module = module
        self.main_class = main_class
        self.inner = inner

    def __eq__(self, other):
        return (
            self.module == other.module
            and self.main_class == other.main_class
            and self.inner == other.inner
        )

    @staticmethod
    def parse_qualified_name(obj: Any) -> Tuple[str, str]:
        qualified_name = obj.__qualname__
        top_level_name, *rest = qualified_name.split(".", 1)

        rest = rest[0] if rest else ""

        return top_level_name, rest

    @staticmethod
    def get_module_and_class(obj: Any) -> Tuple[str, Type[Any] | Empty]:
        """
        Returns (module_name, class/function/type) for `obj`.
        If `obj` is a module, the second item is None.
        """
        if inspect.ismodule(obj):
            return obj.__name__, empty

        module_name = getattr(obj, "__module__", None)
        if module_name:
            if not hasattr(obj, "__qualname__"):
                obj = type(obj)
            return module_name, obj

        obj = type(obj)
        return obj.__module__, obj

    @property
    def full_import_path(self) -> str:
        if self.main_class:
            return f"{self.module}.{self.main_class}"
        return f"{self.module}"

    @property
    def import_path_resolver_data(self) -> str:
        if not self.main_class and not self.full_import_path.startswith("__file__"):
            return f"__module__.{self.full_import_path}"

        return self.full_import_path

    @property
    def qualified_name(self):
        if self.has_inner_inner:
            return f"{self.main_class}.{self.inner}"
        return f"{self.main_class}"

    @property
    def has_inner_inner(self) -> bool:
        return bool(self.inner)

    def get_usage(self, alias):
        if not alias or alias == self.main_class:
            return self.qualified_name

        if self.has_inner_inner:
            return f"{alias}.{self.inner}"

        return alias


class TrieNode:
    """
    A node in the trie, storing a dictionary of children keyed by token.
    """

    __slots__ = ("children",)

    def __init__(self) -> None:
        self.children: Dict[str, "TrieNode"] = {}

    def __repr__(self) -> str:
        return f"{self.children}"

    def insert_child(self, name, child: TrieNode) -> TrieNode:
        self.children.setdefault(name, child)
        return self.children[name]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, TrieNode) and self.children == other.children


class ImportPathResolver:
    """
    A trie for resolving minimal, uniquely identifying import paths.

    Each inserted object or string is converted to a "reversed token" path
    and stored in the trie, so that we can later determine the shortest
    unique suffix of the original import path.
    """

    def __init__(self, *initial_values: Any) -> None:
        self.root = TrieNode()
        self.insert_many(*initial_values)

    def insert(self, value: Any) -> None:
        """
        Insert a value (class, module, or string) into the trie.
        """
        import_path = self.parts_of_import_path(value).import_path_resolver_data

        tokens_reversed = list(reversed(import_path.split(".")))
        current = self.root
        for token in tokens_reversed:
            current = current.insert_child(token, TrieNode())

    def insert_many(self, *values: Any) -> None:
        """
        Insert multiple values into the trie.
        """
        for val in values:
            self.insert(val)

    def get_usage_name(self, value: Any) -> str:
        """
        Extract a minimal, unique suffix (class or last token) for a given
        import path.
        """
        parts = self.parts_of_import_path(value)
        if not parts.module:
            return parts.main_class

        import_path = parts.import_path_resolver_data
        _, suffix = self.find_lcp_parts_for_import(import_path)

        return self._get_alias(suffix, parts)

    @staticmethod
    def _get_alias(suffix, parts):

        alias = "_".join(suffix).strip()
        return parts.get_usage(alias)

    @staticmethod
    def __build_multi_import_statement(
        module,
        imports: List[Tuple[Optional[str], Optional[str]]],
    ) -> str:
        imports = imports or []
        if not imports:
            raise ValueError("No imports specified")

        d = [
            f"{class_name} as {alias}" if alias else class_name
            for class_name, alias in imports
        ]

        joined = f",\n{DEFAULT_INDENTATION}".join(sorted(d))
        return f"from {module} import (\n{DEFAULT_INDENTATION}{joined}\n)"

    @staticmethod
    def __build_single_import_statement(
        module, class_name: Optional[str] = None, alias: Optional[str] = None
    ) -> str:
        statement = f"import {module}"

        if class_name:
            statement = f"from {module} import {class_name}"

        if alias:
            statement = f"{statement} as {alias}"

        return statement

    def build_all_import_statements(self):
        """
        Build import statements for every path that was inserted into the trie,
        by enumerating all leaf nodes. No separate attribute is needed.
        """
        statements = []
        import_parts = defaultdict(list)
        import_data = defaultdict(list)

        self.gather_leaf_paths(self.root, import_parts, (), ())
        for commons, unique_parts in import_parts.items():
            for uniques in unique_parts:
                if not uniques and not commons:
                    continue
                if not commons:
                    *module, obj = uniques
                    module = tuple(module)
                    alias = ()
                else:
                    *module, obj = commons
                    module = tuple(module)
                    alias = (uniques[-1],) + commons
                    module = uniques + module

                if module[0] == "__module__":
                    module = module[1:]

                alias = "_".join(alias).strip().replace("__", "")
                module = ".".join(module).strip()
                if (
                    self.is_builtin_or_keyword(module)
                    or self.is_builtin_or_keyword(obj)
                    or module.startswith("__file__")
                ):
                    continue
                import_data[module].append((obj, alias))
        for module, data in import_data.items():
            if not module:
                for class_name, alias in data:
                    module = class_name
                    class_name = None
                    statement = self.__build_single_import_statement(
                        module, class_name, alias
                    )
                    statements.append(statement)
            elif len(data) == 1:
                statement = self.__build_single_import_statement(module, *data[0])
                statements.append(statement)
            else:
                statement = self.__build_multi_import_statement(module, data)
                statements.append(statement)
        import_statements = [s for s in statements if s.startswith("import")]
        from_statements = [s for s in statements if s.startswith("from")]

        statements = sorted(import_statements) + sorted(from_statements)
        return statements

    def gather_leaf_paths(
        self,
        node: TrieNode,
        imports: DefaultDict,
        common_parts: Tuple = (),
        unique_parts: Tuple = (),
    ) -> DefaultDict[Tuple[str], List[Tuple[str]]]:
        """
        Recursively gather all reversed token paths that end at leaf nodes.
        Each leaf node represents a complete, inserted path.
        """
        if not node.children:
            return imports[common_parts].append(unique_parts)

        for token, child_node in node.children.items():
            uniques = unique_parts
            commons = common_parts

            if len(child_node.children) < 2:
                uniques = (token,) + uniques
            elif uniques:
                commons = (token,) + uniques + common_parts
                uniques = unique_parts = ()
            else:
                commons = (token,) + commons

            self.gather_leaf_paths(child_node, imports, commons, uniques)
        return imports

    @classmethod
    def parts_of_import_path(cls, obj: Any) -> ImportParts:
        """
        Determine the full import path for `obj`. Returns None if the object
        is from a built-in (__main__ or builtins).
        If `obj` is a string, assume it's already a valid import path.
        """
        return ImportParts(obj)

    @classmethod
    def is_builtin_or_keyword(cls, obj: Any) -> bool:
        """
        Check if the given object/string corresponds to the builtins module.
        """
        if isinstance(obj, str):
            return obj in __builtins__ or keyword.iskeyword(obj)

        module_name, _ = ImportParts.get_module_and_class(obj)
        return module_name == "builtins"

    def find_lcp_parts_for_import(
        self, import_path: str
    ) -> Tuple[List[str], List[str]]:
        """
        Split the import path into (prefix, suffix), where suffix is the
        shortest unique import tail determined by the trie.

        If the path is builtin or empty, return ("", path).

        Examples
        --------
        >>> rt = ImportPathResolver()
        >>> rt.insert_many(["a.b.z", "x.y.z"])
        >>> rt.find_lcp_parts_for_import("a.b.z")
        (('a',), ('b', 'z'))
        """
        tokens = import_path.split(".")
        rev_tokens = list(reversed(tokens))
        k = self._find_unique_suffix_length(rev_tokens)

        prefix_tokens = tokens[:-k]
        suffix_tokens = tokens[-k:]

        return prefix_tokens, suffix_tokens

    def _find_unique_suffix_length(self, reversed_tokens: List[str]) -> int:
        """
        Walk down the trie with these reversed tokens. Return the minimal k
        such that after walking k tokens, the current trie node has exactly
        one child. If we never see a node with exactly one child, return
        len(reversed_tokens).
        """
        current = self.root
        for i, token in enumerate(reversed_tokens, start=1):
            current = current.children[token]
            if len(current.children) == 1:
                return i
        return len(reversed_tokens)


def convert_to_class_name(s: str) -> str:
    """
    Convert an arbitrary string into a valid Python class name
    following CamelCase/CapWords convention from PEP 8.

    Examples:
      - "hello world" -> "HelloWorld"
      - "123 invalid!! class" -> "_123InvalidClass"
      - "__some_mixed__Case__" -> "SomeMixedCase"
    """

    if not isinstance(s, str):
        class_name = getattr(s, "__name__", s.__class__.__name__)
        raise ValueError("Invalid argument type: {}".format(class_name))

    s = re.sub(r"[^0-9a-zA-Z]+", " ", s)

    if not s.strip():
        raise ValueError("Class name cannot be empty.")

    words = s.split()
    words = [w.capitalize() for w in words]
    class_name = "".join(words)

    if class_name[0].isdigit():
        class_name = "_" + class_name

    return class_name


def convert_to_attribute_name(s: str) -> str:
    if not isinstance(s, str):
        class_name = getattr(s, "__name__", s.__class__.__name__)
        raise ValueError("Invalid argument type: {}".format(class_name))

    s = re.sub(r"[^0-9a-zA-Z]+", " ", s)

    if not s.strip():
        raise ValueError("Class name cannot be empty.")

    words = s.split()
    attr_name = "_".join(words)

    if attr_name[0].isdigit():
        attr_name = "_" + attr_name

    return attr_name


def create_table_name(table: str, schema: Optional[str] = None) -> str:
    """
    Constructs a fully qualified table name, optionally including the schema.
    """
    table = table.strip()

    if not table:
        raise ValueError("Table name cannot be empty or whitespace.")

    if schema:
        schema = schema.strip()
        return f"{schema}.{table}"

    return table


def create_enum(name, args):
    data = dict(zip(args, args))
    return Enum(name, data)


def to_snake_case(text: str) -> str:
    """
    Converts a given string to snake_case.
    """
    text = re.sub(r"[\-\s]+", "_", text)
    text = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", text)
    text = re.sub(r"([A-Z]+)([A-Z].?)", r"\1_\2", text)
    text = re.sub(r"_+", "_", text)
    return text.lower()


def to_camel_case(text: str) -> str:
    """
    Converts a given string to camelCase.
    """
    if re.search(r"[\-\s_]", text):
        words = re.split(r"[\-\s_]+", text)
        return words[0].lower() + "".join(word.capitalize() for word in words[1:])
    else:
        return text[0].lower() + text[1:] if text else ""


def get_annotation_of_type(type_: Any) -> Any:
    builtin_to_typing = {
        list: List,
        dict: Dict,
        set: Set,
        tuple: Tuple,
        frozenset: FrozenSet,
        deque: Deque,
        defaultdict: DefaultDict,
        ColOrderedDict: OrderedDict,
    }
    return builtin_to_typing.get(type_, type_)


def make_in_file_obj(name: Union[str, Type]) -> str:
    if inspect.isclass(name):
        name = name.__name__
    return f"__file__.{name}"


def resolve_column_type(column_type: TypeEngine) -> Any:
    ColumnTypes = namedtuple(
        "ColumnTypes", ["sql_types", "sql_generic_types", "python_types"]
    )

    if isinstance(column_type, ARRAY):
        column_types = set()
        sql_generic_types = set()
        python_types = set()
        while True:
            column_types.add(column_type.__class__)
            sql_generic_types.add(column_type.as_generic().__class__)
            annotated_type = get_annotation_of_type(column_type.python_type)
            python_types.add(annotated_type)

            if not isinstance(column_type, ARRAY):
                break

            column_type = column_type.item_type
        return ColumnTypes(
            tuple(column_types), tuple(sql_generic_types), tuple(python_types)
        )

    sql_types = (column_type.__class__,)
    sql_generic_types = (column_type.as_generic().__class__,)
    annotated_type = get_annotation_of_type(column_type.python_type)
    python_types = (annotated_type,)

    if isinstance(column_type, SqlEnum):
        python_types = (make_in_file_obj(column_type.name), Enum)

    return ColumnTypes(sql_types, sql_generic_types, python_types)



from collections import defaultdict
from functools import cached_property
from typing import (
    Optional,
    List,
    Type,
    Tuple,
)

from sqlalchemy import (
    Enum as SQLAlchemyEnum,
    Table,
    Column,
    ForeignKey,
    ForeignKeyConstraint,
    Index,
    UniqueConstraint,
    CheckConstraint,
    PrimaryKeyConstraint,
    Computed,
    Identity,
    MetaData,
)
from sqlalchemy.engine.reflection import _ReflectionInfo
from sqlalchemy.orm import (
    DeclarativeBase,
    mapped_column,
    Mapped,
    relationship,
)
from sqlmodel import SQLModel, Field, Relationship

from schema_alchemist.constants import SchemaTypeEnum, SQLRelationshipType
from schema_alchemist.generators import (
    DeclarativeTableGenerator,
    EnumGenerator,
    SQLModelTableGenerator,
    TableGenerator,
)
from schema_alchemist.utils import (
    ImportPathResolver,
    convert_to_class_name,
    inflect_engine,
    to_camel_case,
    to_snake_case,
    make_in_file_obj,
    resolve_column_type,
)


class CoreSchemaGenerator:
    def __init__(
        self,
        reflected_data: _ReflectionInfo,
        sorted_tables_and_fks: List[Tuple[str, List[Tuple[str, str]]]],
        schema: Optional[str] = None,
        add_comments: bool = False,
        create_table_args: bool = False,
        use_camel_case: bool = False,
        use_generic_types: bool = False,
    ):
        self.reflected_data = reflected_data
        self.sorted_tables_and_fks = sorted_tables_and_fks
        self.schema = schema
        self.create_table_args = create_table_args
        self.add_comments = add_comments
        self.use_camel_case = use_camel_case
        self.use_generic_types = use_generic_types

        self.import_path_resolver = ImportPathResolver()

        self.tables = list(self.reflected_data.columns.keys())
        self.table_class_name_map = {
            table: convert_to_class_name(inflect_engine.to_singular(table[1]))
            for table in self.sorted_tables
        }

        self.enum_names = []

    @cached_property
    def sorted_tables(self) -> List[Tuple[Optional[str], str]]:
        sorted_tables = [
            (self.schema, t[0]) for t in self.sorted_tables_and_fks if t[0] is not None
        ]
        views = list(set(self.tables) - set(sorted_tables))
        return sorted_tables + views

    @property
    def metadata_name(self) -> str:
        return self.find_unique_name("metadata")

    def find_unique_name(self, name: str) -> str:
        table_names = list(self.table_class_name_map.values())
        while True:
            if name not in table_names and name not in self.enum_names:
                break
            name += "_"
        return name

    @property
    def schema_type_imports(self):
        return Table, Column, MetaData

    def collect_imports(self):
        meta = make_in_file_obj(self.metadata_name)
        imports = {
            meta,
            *self.schema_type_imports,
        }

        indexes = [
            index
            for indexes in self.reflected_data.indexes.values()
            for index in indexes
            if not index.get("duplicates_constraint")
        ]
        if indexes and any(indexes):
            imports.add(Index)

        checks = [check for check in self.reflected_data.check_constraints.values()]
        if checks and any(checks):
            imports.add(CheckConstraint)

        unique_indexes = [
            index for index in self.reflected_data.unique_constraints.values()
        ]
        if unique_indexes and any(unique_indexes):
            imports.add(UniqueConstraint)

        pks = [pk for pk in self.reflected_data.pk_constraint.values()]
        if pks and any(pks):
            imports.add(PrimaryKeyConstraint)

        fks = [fk for fk in self.reflected_data.foreign_keys.values()]
        if fks and any(fks):
            imports.update([ForeignKey, ForeignKeyConstraint])

        for table in self.sorted_tables:
            table_class_name = self.table_class_name_map[table]
            imports.add(make_in_file_obj(table_class_name))
            for column in self.reflected_data.columns[table]:
                column_type = resolve_column_type(column["type"])

                if column["nullable"]:
                    imports.add(Optional)

                if column.get("computed"):
                    imports.add(Computed)

                if column.get("identity"):
                    imports.add(Identity)

                imports.update(column_type.sql_types)
                imports.update(column_type.python_types)

                if self.use_generic_types:
                    imports.update(column_type.sql_generic_types)

        self.import_path_resolver.insert_many(*imports)

    def generate_base_definition(self) -> str:
        usage = self.import_path_resolver.get_usage_name(MetaData)
        return f"{self.metadata_name} = {usage}(schema={self.schema!r})"

    def generate_imports(self):
        return "\n".join(self.import_path_resolver.build_all_import_statements())

    def generate(self) -> str:
        self.collect_imports()
        tables_generators = [
            TableGenerator(
                name=table[1],
                import_path_resolver=self.import_path_resolver,
                schema=self.schema,
                metadata_name=self.metadata_name,
                columns=self.reflected_data.columns[table],
                comment=self.reflected_data.table_comment[table],
                check_constraints=self.reflected_data.check_constraints[table],
                foreign_keys=self.reflected_data.foreign_keys[table],
                indexes=self.reflected_data.indexes[table],
                primary_key=self.reflected_data.pk_constraint[table],
                unique_constraints=self.reflected_data.unique_constraints[table],
            )
            for table in self.sorted_tables
        ]

        import_statements = self.generate_imports()

        metadata = self.generate_base_definition()

        tables = [tg.generate() for tg in tables_generators]
        return "\n\n\n".join([import_statements, metadata, *tables])


class DeclarativeSchemaGenerator(CoreSchemaGenerator):
    def __init__(
        self,
        reflected_data: _ReflectionInfo,
        sorted_tables_and_fks: List[Tuple[str, List[Tuple[str, str]]]],
        schema: Optional[str] = None,
        add_comments: bool = False,
        create_table_args: bool = False,
        use_camel_case: bool = False,
    ):
        super().__init__(
            reflected_data,
            sorted_tables_and_fks,
            schema,
            add_comments,
            create_table_args,
            use_camel_case,
        )
        self.relationships = defaultdict(list)
        self.relation_names_map = defaultdict(list)
        self.table_column_map = defaultdict(set)
        self.m2m_associated_tables = []

        self.table_class_name_map = {
            table: convert_to_class_name(inflect_engine.to_singular(table[1]))
            for table in self.sorted_tables
        }

        self.table_pk_map = {
            table: pk_data["constrained_columns"]
            for table, pk_data in self.reflected_data.pk_constraint.items()
        }

        self.table_unique_cols_map = defaultdict(list)
        for table, values in self.reflected_data.unique_constraints.items():
            for value in values:
                self.table_unique_cols_map[table].append(value["column_names"])

        self.nullable_column_map = defaultdict(set)
        self.enums = []
        self.columns_types = set()

        for table, columns in self.reflected_data.columns.items():
            for column in columns:
                column_name = column["name"]
                column_type = column["type"]

                self.table_column_map[table].add(column_name)

                if column["nullable"]:
                    self.nullable_column_map[table].add(column_name)

                if isinstance(column_type, SQLAlchemyEnum):
                    table_class = self.table_class_name_map[table]
                    name = column_type.name or f"{table_class}{column_name}"
                    members = column_type.enums
                    self.enums.append((name, members))

                self.columns_types.add(column_type)

    @property
    def metadata_name(self) -> str:
        return self.find_unique_name("Base")

    @property
    def schema_type_imports(self):
        return DeclarativeBase, Mapped, mapped_column, relationship, Column, Table

    def resolve_m2m_relationship(self, table, fks):
        if not fks:
            return None

        pk_columns = set(self.table_pk_map[table])
        column_names = self.table_column_map[table]
        fk_columns = set()
        target_tables = set()
        for fk in fks:
            fk_columns.update({col for col in fk["constrained_columns"]})
            target_tables.add((fk["referred_schema"], fk["referred_table"]))

        if fk_columns == column_names:
            return target_tables

        non_pk_columns = column_names - pk_columns
        if non_pk_columns == fk_columns:
            return target_tables

        return None

    def resolve_relationship_type_of_fk(self, table, fk):
        fk_columns = fk["constrained_columns"]

        if (
            fk_columns in self.table_unique_cols_map[table]
            or fk_columns == self.table_pk_map[table]
        ):
            return SQLRelationshipType.o2o

        return SQLRelationshipType.o2m

    @property
    def singular_suffixes(self) -> List[str]:
        if self.use_camel_case:
            return ["Detail", "Instance", "Data"]
        return ["_detail", "_instance", "_data"]

    @property
    def plural_suffixes(self) -> List[str]:
        if self.use_camel_case:
            return ["Set", "List", "Data"]
        return ["_set", "_list", "_data"]

    def get_suffixes(self, singular: bool = True) -> List[str]:
        if singular:
            return self.singular_suffixes
        return self.plural_suffixes

    def table_has_attribute(self, attribute, table):
        columns = self.table_column_map[table]
        relationships = self.relation_names_map[table]
        return attribute in columns or attribute in relationships

    def find_unique_key_for_relation_attribute(
        self,
        attribute_name,
        main_tabel,
        target_table,
        use_singular_suffixes,
    ) -> str:

        if self.table_has_attribute(attribute_name, main_tabel):
            suffixes = self.get_suffixes(use_singular_suffixes)
            attr_name_singular = inflect_engine.to_singular(attribute_name)
            new_name = None

            for suffix in suffixes:
                tmp_attribute_name = f"{attr_name_singular}{suffix}"

                if not self.table_has_attribute(tmp_attribute_name, main_tabel):
                    new_name = tmp_attribute_name
                    break

            # raise error if no attribute name found
            if new_name is None:
                raise ValueError(
                    "No suitable relationship attribute "
                    "name found for {} in Table: {}".format(
                        target_table[1], main_tabel[1]
                    )
                )

            return new_name

        return attribute_name

    def convert_to_relation_attribute_name(
        self, main_tabel, target_table, relation_type
    ):
        attribute_name = inflect_engine.to_singular(target_table[1].lower())
        use_singular_suffixes = True

        if relation_type in (SQLRelationshipType.m2o, SQLRelationshipType.m2m):
            use_singular_suffixes = False
            attribute_name = inflect_engine.to_plural(attribute_name)

        attribute_name = self.find_unique_key_for_relation_attribute(
            attribute_name,
            main_tabel,
            target_table,
            use_singular_suffixes,
        )

        if self.use_camel_case:
            return to_camel_case(attribute_name)

        return to_snake_case(attribute_name)

    def resolve_relation_and_back_populates_names(
        self, main_table, target_table, relation_type: SQLRelationshipType
    ) -> Tuple[str, str]:

        attribute_name = self.convert_to_relation_attribute_name(
            main_table, target_table, relation_type
        )

        back_populates = self.convert_to_relation_attribute_name(
            target_table, main_table, relation_type.reversed_relationship
        )

        return attribute_name, back_populates

    def create_relationship_metadata(
        self,
        main_table: Tuple[Optional[str], str],
        target_table: Tuple[Optional[str], str],
        relation_type: SQLRelationshipType,
        nullable: bool = False,
        secondary_table: Optional[str] = None,
    ):
        main_class = self.table_class_name_map[main_table]
        target_class = self.table_class_name_map[target_table]
        attribute_name, back_populates = self.resolve_relation_and_back_populates_names(
            main_table, target_table, relation_type
        )

        relationship_data = {
            "attribute_name": attribute_name,
            "target_class": target_class,
            "back_populates": back_populates,
            "relation_type": relation_type,
            "nullable": nullable,
            "secondary_table": secondary_table,
        }
        reverse_relation_data = {
            "attribute_name": back_populates,
            "target_class": main_class,
            "back_populates": attribute_name,
            "relation_type": relation_type.reversed_relationship,
            "nullable": nullable,
            "secondary_table": secondary_table,
        }
        if relationship_data not in self.relationships[main_table]:
            self.relationships[main_table].append(relationship_data)

        if reverse_relation_data not in self.relationships[target_table]:
            self.relationships[target_table].append(reverse_relation_data)

    def handle_m2m_relations(self, secondary_table, tables):
        secondary_class = self.table_class_name_map[secondary_table]
        for table in sorted(tables, key=lambda t: t[1]):
            target_tables = tables - {table}
            for target_table in sorted(target_tables, key=lambda t: t[1]):
                self.create_relationship_metadata(
                    table,
                    target_table,
                    SQLRelationshipType.m2m,
                    nullable=False,
                    secondary_table=secondary_class,
                )

    def resolve_relationships(self):

        for main_table, fks in self.reflected_data.foreign_keys.items():
            target_tables = self.resolve_m2m_relationship(main_table, fks)

            if target_tables:
                self.m2m_associated_tables.append(main_table)
                self.handle_m2m_relations(main_table, target_tables)
                continue

            for fk in fks:
                target_table = (fk["referred_schema"], fk["referred_table"])
                relation_type = self.resolve_relationship_type_of_fk(main_table, fk)

                fk_columns = fk["constrained_columns"]
                nullable = set(fk_columns).issubset(
                    set(self.nullable_column_map[main_table])
                )

                self.create_relationship_metadata(
                    main_table,
                    target_table,
                    relation_type,
                    nullable,
                    None,
                )

    def generate_base_definition(self) -> str:
        declarative_class = self.import_path_resolver.get_usage_name(DeclarativeBase)
        return f"class {self.metadata_name}({declarative_class}):\n    pass"

    def generate_enums(self):
        return [
            EnumGenerator(name, items, self.import_path_resolver).generate()
            for name, items in self.enums
        ]

    def generate(self) -> str:
        self.collect_imports()
        columns = self.reflected_data.columns
        table_comment = self.reflected_data.table_comment
        check_constraints = self.reflected_data.check_constraints
        foreign_keys = self.reflected_data.foreign_keys
        indexes = self.reflected_data.indexes
        pk_constraint = self.reflected_data.pk_constraint
        unique_constraints = self.reflected_data.unique_constraints

        self.resolve_relationships()

        enums = self.generate_enums()

        tables_generators = []
        for table in reversed(self.sorted_tables):
            if table in self.m2m_associated_tables:
                tables_generators.append(
                    TableGenerator(
                        name=table[1],
                        import_path_resolver=self.import_path_resolver,
                        schema=self.schema,
                        metadata_name=f"{self.metadata_name}.metadata",
                        columns=columns[table],
                        comment=table_comment[table],
                        check_constraints=check_constraints[table],
                        foreign_keys=foreign_keys[table],
                        indexes=indexes[table],
                        primary_key=pk_constraint[table],
                        unique_constraints=unique_constraints[table],
                    )
                )

            else:
                generator = DeclarativeTableGenerator(
                    name=table[1],
                    import_path_resolver=self.import_path_resolver,
                    schema=self.schema,
                    metadata_name=self.metadata_name,
                    columns=columns[table],
                    comment=table_comment[table],
                    check_constraints=check_constraints[table],
                    foreign_keys=foreign_keys[table],
                    indexes=indexes[table],
                    primary_key=pk_constraint[table],
                    unique_constraints=unique_constraints[table],
                    relationships=self.relationships[table],
                )
                tables_generators.append(generator)

        import_statements = self.generate_imports()

        tables = [tg.generate() for tg in tables_generators]
        return "\n\n\n".join(
            [import_statements, *enums, self.generate_base_definition(), *tables]
        )


class SQLModelSchemaGenerator(DeclarativeSchemaGenerator):

    @cached_property
    def schema_type_imports(self):
        return SQLModel, Field, Relationship, Column

    def generate_base_definition(self) -> str:
        return ""

    def generate(self):
        self.collect_imports()
        self.resolve_relationships()
        enums = self.generate_enums()
        tables_generators = [
            SQLModelTableGenerator(
                name=table[1],
                import_path_resolver=self.import_path_resolver,
                schema=self.schema,
                metadata_name=self.metadata_name,
                columns=self.reflected_data.columns[table],
                comment=self.reflected_data.table_comment[table],
                check_constraints=self.reflected_data.check_constraints[table],
                foreign_keys=self.reflected_data.foreign_keys[table],
                indexes=self.reflected_data.indexes[table],
                primary_key=self.reflected_data.pk_constraint[table],
                unique_constraints=self.reflected_data.unique_constraints[table],
                relationships=self.relationships[table],
            )
            for table in reversed(self.sorted_tables)
        ]

        import_statements = self.generate_imports()
        tables = [tg.generate() for tg in tables_generators]
        return "\n\n\n".join([import_statements, *enums, *tables])


class SchemaGeneratorFactory:
    def __init__(
        self,
        reflected_data: _ReflectionInfo,
        sorted_tables_and_fks: List[Tuple[str, List[Tuple[str, str]]]],
        schema_type: SchemaTypeEnum = SchemaTypeEnum.table,
        schema: Optional[str] = None,
        add_comments: bool = False,
        create_table_args: bool = False,
        use_camel_case: bool = False,
    ):
        self.reflected_data = reflected_data
        self.sorted_tables_and_fks = sorted_tables_and_fks
        self.schema = schema
        self.add_comments = add_comments
        self.create_table_args = create_table_args
        self.use_camel_case = use_camel_case
        self.schema_type = schema_type

    def get_generator_class(self) -> Type[CoreSchemaGenerator]:
        if self.schema_type == SchemaTypeEnum.declarative:
            return DeclarativeSchemaGenerator
        if self.schema_type == SchemaTypeEnum.sqlmodel:
            return SQLModelSchemaGenerator
        if self.schema_type == SchemaTypeEnum.table:
            return CoreSchemaGenerator
        raise ValueError(f"Unknown schema type: {self.schema_type}")

    def generate(self) -> str:
        generator_class = self.get_generator_class()

        generator = generator_class(
            self.reflected_data,
            self.sorted_tables_and_fks,
            self.schema,
            self.add_comments,
            self.create_table_args,
            self.use_camel_case,
        )
        return generator.generate()



import pytest

from schema_alchemist.constants import SQLRelationshipType
from schema_alchemist.generators import DeclarativeRelationGenerator


@pytest.mark.parametrize(
    "data, expected",
    (
        (
            {
                "attribute_name": "user",
                "target_class": "User",
                "back_populates": "profile",
                "nullable": False,
                "secondary": None,
                "relation_type": SQLRelationshipType.o2o,
            },
            "    user: Mapped['User'] = relationship(back_populates='profile')",
        ),
        (
            {
                "attribute_name": "parent",
                "target_class": "Parent",
                "back_populates": "children",
                "nullable": False,
                "secondary": None,
                "relation_type": SQLRelationshipType.o2m,
            },
            "    parent: Mapped['Parent'] = relationship(" "back_populates='children')",
        ),
        (
            {
                "attribute_name": "children",
                "target_class": "Child",
                "back_populates": "parent",
                "nullable": False,
                "secondary": None,
                "relation_type": SQLRelationshipType.m2o,
            },
            "    children: Mapped[List['Child']] = relationship("
            "back_populates='parent')",
        ),
        (
            {
                "attribute_name": "attributes",
                "target_class": "Attribute",
                "back_populates": "categories",
                "nullable": False,
                "secondary": None,
                "secondary_table": "CategoryAttribute",
                "relation_type": SQLRelationshipType.m2m,
            },
            "    attributes: Mapped[List['Attribute']] = relationship("
            "secondary=CategoryAttribute, back_populates='categories')",
        ),
    ),
)
def test_declarative_relation_generator(data, expected, pre_configured_ipr):
    generator = DeclarativeRelationGenerator(
        import_path_resolver=pre_configured_ipr, **data
    )

    assert expected == generator.generate()



from typing import (
    Optional,
    List,
    Any,
    Dict,
    Callable,
)

from sqlalchemy.orm import Mapped, relationship
from sqlmodel import Relationship

from schema_alchemist.constants import SQLRelationshipType
from schema_alchemist.generators.base_generators import BaseGenerator
from schema_alchemist.utils import ImportPathResolver, StringReprWrapper


class DeclarativeRelationGenerator(BaseGenerator):

    def __init__(
        self,
        attribute_name: str,
        target_class: str,
        back_populates: str,
        relation_type: SQLRelationshipType,
        import_path_resolver: ImportPathResolver,
        nullable: bool = False,
        secondary_table: Any = None,
        *args,
        **kwargs,
    ):
        super().__init__(import_path_resolver, *args, **kwargs)
        self.attribute_name = attribute_name
        self.target_class = target_class
        self.back_populates = back_populates
        self.nullable = nullable
        self.secondary = secondary_table
        self.relation_type = relation_type

    @property
    def klass(self) -> Callable:
        return relationship

    @property
    def python_annotation(self) -> str:
        annotation = f"'{self.target_class}'"

        if self.nullable:
            optional = self.import_path_resolver.get_usage_name(Optional)
            annotation = f"{optional}[{annotation}]"

        if self.relation_type in (SQLRelationshipType.m2o, SQLRelationshipType.m2m):
            list_usage = self.import_path_resolver.get_usage_name(List)
            annotation = f"{list_usage}[{annotation}]"

        return annotation

    @property
    def parameters(self) -> Dict[str, Any]:
        secondary = self.secondary
        if secondary:
            secondary = StringReprWrapper(secondary)

        return {
            "secondary": secondary,
            "back_populates": self.back_populates,
        }

    def generate_relation(self):
        return self.generate_function_definition(self.klass, self.parameters)

    def generate(self, *args, **kwargs) -> str:
        mapped_import_name = self.import_path_resolver.get_usage_name(Mapped)
        return (
            f"{self.indent}{self.attribute_name}: "
            f"{mapped_import_name}[{self.python_annotation}] = "
            f"{self.generate_relation()}"
        )


class SQLModelRelationGenerator(DeclarativeRelationGenerator):

    @property
    def parameters(self) -> Dict[str, Any]:
        secondary = self.secondary
        if secondary:
            secondary = StringReprWrapper(secondary)
        return {
            "back_populates": self.back_populates,
            "link_model": secondary,
        }

    @property
    def klass(self) -> Callable:
        return Relationship

    def generate(self, *args, **kwargs) -> str:
        return (
            f"{self.indent}{self.attribute_name}: {self.python_annotation} = "
            f"{self.generate_relation()}"
        )



from datetime import datetime, date
from decimal import Decimal

import pytest
from sqlalchemy import (
    Integer,
    VARCHAR,
    NUMERIC,
    DOUBLE_PRECISION,
    DATE,
    INTEGER,
    TEXT,
    TIMESTAMP,
)
from sqlalchemy.dialects.mysql import ENUM

from schema_alchemist.generators import ColumnGenerator
from schema_alchemist.utils import StringReprWrapper


def test_create_fk_constraint(pre_configured_ipr):
    fk_data = {
        "name": "orders_user_id_fkey",
        "constrained_columns": ["user_id"],
        "referred_schema": "public",
        "referred_table": "users",
        "referred_columns": ["id"],
        "options": {"ondelete": "CASCADE"},
        "comment": None,
    }

    expected = StringReprWrapper(
        "ForeignKey(column='users.id', name='orders_user_id_fkey', ondelete='CASCADE')"
    )

    column_data = {
        "type": Integer(),
        "name": "id",
        "nullable": False,
        "foreign_key": fk_data,
    }
    column_generator = ColumnGenerator(column_data, pre_configured_ipr, "TestTable")
    result = column_generator.create_fk_constraint()
    assert result == expected


@pytest.mark.parametrize(
    "input_value, expected",
    (
        (
            {
                "sqltext": "((quantity)::numeric * unit_price)",
                "persisted": True,
            },
            "Computed(sqltext='((quantity)::numeric * unit_price)', persisted=True)",
        ),
        (
            {
                "sqltext": "((quantity)::numeric * unit_price)",
                "persisted": False,
            },
            "Computed(sqltext='((quantity)::numeric * unit_price)', persisted=False)",
        ),
    ),
)
def test_create_column_computed(pre_configured_ipr, input_value, expected):
    column_data = {
        "name": "total_price",
        "type": NUMERIC(precision=10, scale=2),
        "nullable": True,
        "default": None,
        "autoincrement": False,
        "comment": None,
        "computed": input_value,
    }
    expected = StringReprWrapper(expected)
    column_generator = ColumnGenerator(column_data, pre_configured_ipr, "TestTable")
    result = column_generator.create_column_computed()
    assert result == expected


def test_create_column_computed_none(import_path_resolver):
    column_data = {
        "name": "total_price",
        "type": NUMERIC(precision=10, scale=2),
        "nullable": True,
        "default": None,
        "autoincrement": False,
        "comment": None,
    }

    column_generator = ColumnGenerator(column_data, import_path_resolver, "TestTable")
    result = column_generator.create_column_computed()
    assert result is None


@pytest.mark.parametrize(
    "input_value, expected",
    (
        (
            {
                "name": "id",
                "type": INTEGER(),
                "nullable": False,
                "default": "nextval('\"public\".addresses_id_seq'::regclass)",
                "autoincrement": True,
                "comment": None,
            },
            "    Column('id', INTEGER(), autoincrement=True, nullable=False, "
            "server_default='nextval(\\'\"public\".addresses_id_seq\\'::regclass)')",
        ),
        (
            {
                "name": "user_id",
                "type": INTEGER(),
                "nullable": False,
                "default": None,
                "autoincrement": False,
                "comment": None,
            },
            "    Column('user_id', INTEGER(), autoincrement=False, nullable=False)",
        ),
        (
            {
                "name": "address_line1",
                "type": VARCHAR(length=200),
                "nullable": False,
                "default": None,
                "autoincrement": False,
                "comment": None,
            },
            "    Column('address_line1', VARCHAR(length=200), "
            "autoincrement=False, nullable=False)",
        ),
        (
            {
                "name": "created_at",
                "type": TIMESTAMP(),
                "nullable": False,
                "default": "CURRENT_TIMESTAMP",
                "autoincrement": False,
                "comment": None,
            },
            "    Column('created_at', TIMESTAMP(), autoincrement=False, nullable=False, "
            "server_default='CURRENT_TIMESTAMP')",
        ),
        (
            {
                "name": "description",
                "type": TEXT(),
                "nullable": True,
                "default": None,
                "autoincrement": False,
                "comment": None,
            },
            "    Column('description', TEXT(), autoincrement=False, nullable=True)",
        ),
        (
            {
                "name": "order_date",
                "type": DATE(),
                "nullable": True,
                "default": "CURRENT_DATE",
                "autoincrement": False,
                "comment": None,
            },
            "    Column('order_date', DATE(), autoincrement=False, nullable=True, "
            "server_default='CURRENT_DATE')",
        ),
        (
            {
                "name": "status",
                "type": ENUM(
                    "pending",
                    "processed",
                    "shipped",
                    "delivered",
                    "cancelled",
                    name="order_status_enum",
                ),
                "nullable": False,
                "default": "'pending'::order_status_enum",
                "autoincrement": False,
                "comment": None,
            },
            "    Column('status', ENUM('pending', 'processed', 'shipped', 'delivered', "
            "'cancelled'), autoincrement=False, nullable=False, "
            "server_default=\"'pending'::order_status_enum\")",
        ),
        (
            {
                "name": "total_amount",
                "type": NUMERIC(precision=10, scale=2),
                "nullable": False,
                "default": None,
                "autoincrement": False,
                "comment": None,
            },
            "    Column('total_amount', NUMERIC(precision=10, scale=2), "
            "autoincrement=False, nullable=False)",
        ),
        (
            {
                "name": "unit_price",
                "type": DOUBLE_PRECISION(precision=53),
                "nullable": False,
                "default": None,
                "autoincrement": False,
                "comment": None,
            },
            "    Column('unit_price', DOUBLE_PRECISION(precision=53), "
            "autoincrement=False, nullable=False)",
        ),
        (
            {
                "name": "total_price",
                "type": NUMERIC(precision=10, scale=2),
                "nullable": True,
                "default": None,
                "autoincrement": False,
                "comment": None,
                "computed": {
                    "sqltext": "((quantity)::numeric * unit_price)",
                    "persisted": True,
                },
            },
            "    Column('total_price', NUMERIC(precision=10, scale=2), "
            "Computed(sqltext='((quantity)::numeric * unit_price)', persisted=True), "
            "autoincrement=False, nullable=True)",
        ),
        (
            {
                "name": "id",
                "type": INTEGER(),
                "nullable": False,
                "default": None,
                "autoincrement": True,
                "comment": None,
                "identity": {
                    "always": False,
                    "start": 1,
                    "increment": 1,
                    "minvalue": 1,
                    "maxvalue": 2147483647,
                    "cache": 1,
                    "cycle": False,
                },
            },
            "    Column('id', INTEGER(), "
            "Identity(start=1, increment=1, minvalue=1, maxvalue=2147483647, "
            "cycle=False, cache=1), "
            "autoincrement=True, nullable=False)",
        ),
    ),
)
def test_core_column_generate(pre_configured_ipr, input_value, expected):
    column_generator = ColumnGenerator(input_value, pre_configured_ipr, "TestTable")
    result = column_generator.generate()
    assert result == expected


@pytest.mark.parametrize(
    "input_value, expected",
    (
        (
            {
                "name": "id",
                "type": INTEGER(),
                "nullable": False,
            },
            {
                "column_name": "id",
                "column_type": "INTEGER()",
                "column_nullable": False,
                "column_type_class": INTEGER,
                "column_python_type": int,
            },
        ),
        (
            {
                "name": "address",
                "type": VARCHAR(length=200),
                "nullable": False,
            },
            {
                "column_name": "address",
                "column_type": "VARCHAR(length=200)",
                "column_nullable": False,
                "column_type_class": VARCHAR,
                "column_python_type": str,
            },
        ),
        (
            {
                "name": "created_at",
                "type": TIMESTAMP(),
                "nullable": False,
            },
            {
                "column_name": "created_at",
                "column_type": "TIMESTAMP()",
                "column_nullable": False,
                "column_type_class": TIMESTAMP,
                "column_python_type": datetime,
            },
        ),
        (
            {
                "name": "description",
                "type": TEXT(),
                "nullable": True,
            },
            {
                "column_name": "description",
                "column_type": "TEXT()",
                "column_nullable": True,
                "column_type_class": TEXT,
                "column_python_type": str,
            },
        ),
        (
            {
                "name": "order_date",
                "type": DATE(),
                "nullable": True,
            },
            {
                "column_name": "order_date",
                "column_type": "DATE()",
                "column_nullable": True,
                "column_type_class": DATE,
                "column_python_type": date,
            },
        ),
        (
            {
                "name": "total_amount",
                "type": NUMERIC(precision=10, scale=2),
                "nullable": False,
            },
            {
                "column_name": "total_amount",
                "column_type": "NUMERIC(precision=10, scale=2)",
                "column_nullable": False,
                "column_type_class": NUMERIC,
                "column_python_type": Decimal,
            },
        ),
        (
            {
                "name": "unit_price",
                "type": DOUBLE_PRECISION(precision=53),
                "nullable": False,
            },
            {
                "column_name": "unit_price",
                "column_type": "DOUBLE_PRECISION(precision=53)",
                "column_nullable": False,
                "column_type_class": DOUBLE_PRECISION,
                "column_python_type": float,
            },
        ),
    ),
)
def test_properties(pre_configured_ipr, input_value, expected):
    cg = ColumnGenerator(input_value, pre_configured_ipr, "TestTable")
    assert cg.column_name == expected["column_name"]
    assert cg.column_type.__repr__() == expected["column_type"]
    assert cg.column_nullable == expected["column_nullable"]
    assert cg.column_type_class == expected["column_type_class"]
    assert cg.column_python_type == expected["column_python_type"]


def test_column_enum_type(import_path_resolver):
    data = {
        "name": "status",
        "type": ENUM(
            "pending",
            "processed",
            "shipped",
            "delivered",
            "cancelled",
            name="order_status_enum",
        ),
        "nullable": False,
    }

    cg = ColumnGenerator(data, import_path_resolver, "TestTable")

    assert cg.column_python_type == "order_status_enum"

    assert cg.column_name == "status"
    assert (
        cg.column_type.__repr__()
        == "ENUM('pending', 'processed', 'shipped', 'delivered', 'cancelled')"
    )
    assert not cg.column_nullable
    assert cg.column_type_class == ENUM
